import cv from "@yizhi/cv";
import { ImageCropOption, ImageSource, Model } from "../common/model";
import { convertImage } from "../common/processors";
import { FaceAlignmentResult, FacePoint, indexFromTo } from "./common";

interface FaceLandmark1000PredictOption extends ImageCropOption { }

class FaceLandmark1000Result extends FaceAlignmentResult {
	public constructor(points: FacePoint[]) {
		super(points, {
			directionPointIndex: [401, 529],
			leftEyePointIndex: indexFromTo(401, 528),
			rightEyePointIndex: indexFromTo(529, 656),
			leftEyebrowPointIndex: indexFromTo(273, 336),
			rightEyebrowPointIndex: indexFromTo(337, 400),
			mouthPointIndex: indexFromTo(845, 972),
			nosePointIndex: indexFromTo(657, 844),
			contourPointIndex: indexFromTo(0, 272),
		});
	}

}


const MODEL_URL_CONFIG = {
	FACELANDMARK1000_ONNX: `https://www.modelscope.cn/models/luyizhi/basic_cv/resolve/master/facealign/FaceLandmark1000.onnx`,
	FACELANDMARK1000_MNN: `https://www.modelscope.cn/models/luyizhi/basic_cv/resolve/master/facealign/FaceLandmark1000.mnn`,
};

export class FaceLandmark1000 extends Model {

	public static async load(type?: keyof typeof MODEL_URL_CONFIG) {
		return this.cacheModel(MODEL_URL_CONFIG[type ?? "FACELANDMARK1000_ONNX"], { createModel: true }).then(r => r.model);
	}

	public predict(image: ImageSource, option?: FaceLandmark1000PredictOption) { return Model.resolveImage(image, im => this.doPredict(im, option)); }

	public async doPredict(image: cv.Mat, option?: FaceLandmark1000PredictOption) {
		const input = this.input;
		if (option?.crop) image = cv.crop(image, option?.crop);
		const ratioWidth = image.cols / input.shape[3];
		const ratioHeight = image.rows / input.shape[2];
		image = cv.resize(image, input.shape[3], input.shape[2]);

		const nchwImageData = convertImage(image.data, { sourceImageFormat: "bgr", targetColorFormat: "gray", targetShapeFormat: "nchw", targetNormalize: { mean: [0], std: [1] } });

		const res = await this.session.run({
			[input.name]: {
				shape: [1, 1, input.shape[2], input.shape[3]],
				data: nchwImageData,
				type: "float32",
			}
		}).then(res => res[this.output.name].data);

		const points: FacePoint[] = [];
		for (let i = 0; i < res.length; i += 2) {
			const x = res[i] * image.cols * ratioWidth;
			const y = res[i + 1] * image.rows * ratioHeight;
			points.push({ x, y });
		}

		return new FaceLandmark1000Result(points);

	}

}